import sympy as sp
from functools import reduce
from itertools import combinations
from collections import defaultdict
from typing import List, Tuple
import pandas as pd


class Quiver:
    def __init__(self, 
                 vertices: List[int] = [], 
                 clusters: List[Tuple[int, int]] = [],
                 laminations: List[List[Tuple[int, int]]] = [],
                 is_principal: bool = False
                 ):
        self.n = len(vertices)
        self.vertices = vertices
        self.frozens = [(i, (i+1)%self.n) for i in range(self.n)]
        self.clusters = clusters
        self.cluster_names = [str(i) for i in range(1, len(clusters)+1)]
        
        if is_principal:
            laminations = [
                [(self.get_previous_vertex(p), self.get_previous_vertex(q))]
                for p, q in self.clusters
            ] + laminations
        
        self.laminations = laminations
        self.lamination_names = [str(i) for i in range(1, len(laminations)+1)]
        self.x = [
            [sp.symbols(f"""x_{{{i}{"'"*j}}}""") for j in range(self.n)]
            for i in range(1, len(clusters)+1)
        ]
        self.u = [
            sp.symbols(f"""u_{{{i}}}""") for i in range(1, len(laminations)+1)
        ]

    def get_adjacent_edges(self, edge: Tuple[int, int]):
        p, q = edge
        r = []
        all_edges = self.frozens + self.clusters
        for i in range(self.n):
            if (p, i) in all_edges or (i, p) in all_edges:
                if (q, i) in all_edges or (i, q) in all_edges:
                    r.append(i)
        return r


    def get_next_vertex(self, k: int):
        return (k+1)%self.n
    
    def get_previous_vertex(self, k: int):
        return (k-1)%self.n

    
    def flip(self, edge: Tuple[int, int]):
        if edge not in self.clusters:
            raise ValueError("Edge is not a cluster edge")
        idx = self.clusters.index(edge)
        p, q = edge
        r = self.get_adjacent_edges(edge)
        self.clusters[idx] = (r[0], r[1])
        self.cluster_names[idx] += "'"
        return self
    
    def is_in_frozens(self, edge: Tuple[int, int]):
        p, q = edge
        return (p, q) in self.frozens or (q, p) in self.frozens
    
    def is_in_clusters(self, edge: Tuple[int, int]):
        p, q = edge
        return (p, q) in self.clusters or (q, p) in self.clusters
        
    def is_in_edges(self, edge: Tuple[int, int]):
        return self.is_in_frozens(edge) or self.is_in_clusters(edge)
    
    def get_cluster_index_by_edge(self, edge: Tuple[int, int]):
        p, q = edge
        if (p, q) in self.clusters:
            return self.clusters.index((p, q))
        elif (q, p) in self.clusters:
            return self.clusters.index((q, p))
        else:
            raise ValueError("Edge is not a cluster edge")

    def get_all_triangles(self):
        for a, b, c in combinations(range(self.n), 3):
            if self.is_in_edges((a, b)) and self.is_in_edges((b, c)) and self.is_in_edges((c, a)):
                yield (a, b, c)
    
    def get_intersecting_clusters(self, start: int, end: int):
        s, e = start, end
        s += 0.5
        e += 0.5
        
        def between(a, b, c):
            return a < b < c or c < b < a
        
        intersecting_clusters = []
        for p, q in self.clusters:
            if between(s, p, e) ^ between(s, q, e):
                intersecting_clusters.append((p, q))
        
        intersecting_clusters.sort(key=lambda x: ((x[0]-start)%self.n, -(x[1]-start)%self.n))
        return intersecting_clusters
    
    def get_shear_coordinates(self, start: int, end: int):
        intersecting_clusters = []
        intersecting_clusters.append((start, self.get_next_vertex(start)))
        intersecting_clusters.extend(self.get_intersecting_clusters(start, end))
        intersecting_clusters.append((end, self.get_next_vertex(end)))
        
        c = defaultdict(int)
        for p, q in intersecting_clusters:
            c[p] += 1
            c[q] += 1
        
        r = []
        r.append((start, self.get_next_vertex(start)))
        r += [(p, q) for p, q in intersecting_clusters if c[p] >= 2 and c[q] >= 2]
        r.append((end, self.get_next_vertex(end)))
        
        if len(r) <= 2:
            return []
        
        if (c[start] >= 2 and start in r[1]) or (c[end] >= 2 and end in r[1]):
            return [(p, q, (-1)**(i+1)) for i, (p, q) in enumerate(r[1:-1])]
        else:
            return [(p, q, (-1)**i) for i, (p, q) in enumerate(r[1:-1])]

    
    def get_exchange_matrix(self):
        n = len(self.clusters)  # Number of mutable vertices
        m = len(self.laminations)
        
        # Initialize empty matrix with n+m rows and n columns
        matrix = [[0 for _ in range(n)] for _ in range(n+m)]
        
        for a, b, c in self.get_all_triangles():
            if self.is_in_clusters((a, b)) and self.is_in_clusters((b, c)):
                idx1 = self.get_cluster_index_by_edge((a, b))
                idx2 = self.get_cluster_index_by_edge((b, c))
                matrix[idx1][idx2] -= 1
                matrix[idx2][idx1] += 1
            if self.is_in_clusters((b, c)) and self.is_in_clusters((c, a)):
                idx1 = self.get_cluster_index_by_edge((b, c))
                idx2 = self.get_cluster_index_by_edge((c, a))
                matrix[idx1][idx2] -= 1
                matrix[idx2][idx1] += 1
            if self.is_in_clusters((c, a)) and self.is_in_clusters((a, b)):
                idx1 = self.get_cluster_index_by_edge((c, a))
                idx2 = self.get_cluster_index_by_edge((a, b))
                matrix[idx1][idx2] -= 1
                matrix[idx2][idx1] += 1
        
        for i, lamination in enumerate(self.laminations):
            for x, y in lamination:
                for p, q, sign in self.get_shear_coordinates(x, y):
                    matrix[n+i][self.get_cluster_index_by_edge((p, q))] = sign
            
        
        
        # Create column index for clusters
        col_index = [f'x_{i}' for i in self.cluster_names]
        
        # Create row index for clusters and laminations
        row_index = [f'x_{i}' for i in self.cluster_names] + [f'u_{i}' for i in self.lamination_names]
        
        return pd.DataFrame(matrix, index=row_index, columns=col_index)
    
    def flip(self, 
             edge: Tuple[int, int] = None,
             index: int = None):
        if index is not None:
            edge = self.clusters[index-1]
        elif not self.is_in_clusters(edge):
            raise ValueError("Edge is not a cluster edge")
        
        p, q = self.get_adjacent_edges(edge)
        x, y = edge
        
        idx = self.get_cluster_index_by_edge(edge)
        self.clusters[idx] = (p, q)
        self.cluster_names[idx] += "'"
        return self
    
    def get_var_index(self, name: str):
        if name.startswith("x_"):
            base = int(name.split('_')[1].rstrip("'"))
            primes = name.count("'")
            return base-1, primes
        elif name.startswith("u_"):
            return int(name.split('_')[1])-1
        else:
            raise ValueError("Variable name is not valid")
    
    def get_var_by_rowname(self, rowname: str):
        if rowname.startswith("x_"):
            # Extract base number and count primes
            base = int(rowname.split('_')[1].rstrip("'"))
            primes = rowname.count("'")
            return self.x[base-1][primes]
        elif rowname.startswith("u_"):
            return self.u[int(rowname.split('_')[1])-1]
        else:
            raise ValueError("Column name is not valid")
    
    
    def express_target(self, start: int, end: int):
        from copy import deepcopy
        q = deepcopy(self)
        
        relations = []
        expressions = []
        
        for e in q.get_intersecting_clusters(start, end):
            B = q.get_exchange_matrix()
            
            idx = q.get_cluster_index_by_edge(e)
            col = B[f'x_{q.cluster_names[idx]}']
            pos_vars = [q.get_var_by_rowname(B.index[i]) for i,v in enumerate(col) if v == 1]
            neg_vars = [q.get_var_by_rowname(B.index[i]) for i,v in enumerate(col) if v == -1]
            
            pos_term = reduce(lambda x, y: x*y, pos_vars, 1)
            neg_term = reduce(lambda x, y: x*y, neg_vars, 1)
            
            base, primes = q.get_var_index(B.columns[idx])
            
            relations.append(sp.Eq(q.x[base][primes] * q.x[base][primes+1], pos_term + neg_term))
            expressions.append(sp.Eq(q.x[base][primes+1], (pos_term + neg_term) / q.x[base][primes]))
            
            q.flip(edge=e)
        q.relations = relations
        q.expressions = expressions
        
        for i in range(len(q.expressions)-1):
            for j in range(i+1, len(q.expressions)):
                q.expressions[j] = q.expressions[j].subs(q.expressions[i].lhs, q.expressions[i].rhs).expand().simplify()
        
        # Expand only numerators of expressions
        for i in range(len(q.expressions)):
            expr = q.expressions[i]
            if isinstance(expr.rhs, sp.Mul):
                num, den = expr.rhs.as_numer_denom()
                q.expressions[i] = sp.Eq(expr.lhs, (num.expand() / den))

        return q



import matplotlib.pyplot as plt
import numpy as np
from .quiver import Quiver

lamination_palette = [
    '#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
    '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf',
    '#aec7e8', '#ffbb78', '#98df8a', '#ff9896', '#c5b0d5',
    '#c49c94', '#f7b6d2', '#c7c7c7', '#dbdb8d', '#9edae5',
    '#393b79', '#637939', '#8c6d31', '#843c39', '#7b4173',
    '#5254a3', '#8ca252', '#bd9e39', '#ad494a', '#a55194',
    '#6b6ecf', '#b5cf6b', '#e7ba52', '#d6616b', '#ce6dbd',
    '#9c9ede', '#cedb9c', '#e7cb94', '#e7969c', '#de9ed6',
    '#3182bd', '#31a354', '#756bb1', '#636363', '#e6550d',
    '#80b1d3', '#8dd3c7', '#ffffb3', '#bebada', '#fb8072',
    '#6baed6', '#74c476', '#9e9ac8', '#969696', '#fd8d3c',
    '#b3cde3', '#ccebc5', '#decbe4', '#fed9a6', '#ffffcc',
    '#e5d8bd', '#fddaec', '#f2f2f2', '#b3e2cd', '#fdcdac',
    '#a6cee3', '#1f78b4', '#b2df8a', '#33a02c', '#fb9a99',
    '#e31a1c', '#fdbf6f', '#ff7f00', '#cab2d6', '#6a3d9a',
    '#b15928', '#8dd3c7', '#ffffb3', '#bebada', '#fb8072',
    '#80b1d3', '#fdb462', '#b3de69', '#fccde5', '#d9d9d9',
    '#bc80bd', '#ccebc5', '#ffed6f', '#66c2a5', '#fc8d62',
    '#8da0cb', '#e78ac3', '#a6d854', '#ffd92f', '#e5c494',
    '#b3b3b3', '#e41a1c', '#377eb8', '#4daf4a', '#984ea3'
]

class Plotter:
    def __init__(self):
        pass
    
    def plot_quiver(self, 
                   quiver: Quiver,
                   ax: plt.Axes = None,
                   vertex_labels: bool = True,
                   frozen_labels: bool = True,
                   cluster_labels: bool = True,
                   plot_laminations: bool = True,
                   laminations_labels: bool = True,
                   rotation: float = 0, 
                   fontsize: int = 7,
                   lamination_label_pos: float = 0.5,
                   ):
        n = quiver.n
        # Calculate polygon vertices positions
        angles = np.linspace(0, 2*np.pi, n, endpoint=False)
        radius = 10
        x = radius * np.cos(angles + rotation)
        y = radius * np.sin(angles + rotation)
        
        # Create figure if not provided
        if ax is None:
            _, ax = plt.subplots()
            ax.set_xticks([])
            ax.set_yticks([])
            ax.axis('off')
        
        
        ############################ POLYGON ############################
        # Plot polygon
        polygon = plt.Polygon(np.column_stack([x, y]), fill=False, linewidth=0.2)
        ax.add_patch(polygon)
        
        #  VERTEX LABELS
        if vertex_labels:
            for i in range(n):
                # Add white circle background for label
                center_x = x[i]
                center_y = y[i]
                circle = plt.Circle((center_x, center_y), 0.1, color='white')
                ax.add_patch(circle)
                
                # Add text label
                ax.text(center_x, center_y, f'v_{i}', 
                        horizontalalignment='center',
                        verticalalignment='center', fontsize=fontsize)
        
        ############################ FROZEN EDGES ############################
        for i in range(len(quiver.frozens)):
            p, q = quiver.frozens[i]
            
            center_x = (x[p] + x[(p+1)%n])/2
            center_y = (y[p] + y[(p+1)%n])/2
            
            ax.plot([x[p], x[q]], [y[p], y[q]], color='black')
            
            if frozen_labels:
                # Add white circle background for label
                circle = plt.Circle((center_x, center_y), 0.1, color='white')
                ax.add_patch(circle)
                
                # Add text label
                ax.text(center_x, center_y, f'e_{i}',
                        horizontalalignment='center',
                        verticalalignment='center', fontsize=fontsize)
        
        ############################ CLUSTERS ############################
        for i in range(len(quiver.clusters)):
            p, q = quiver.clusters[i]
            
            ax.plot([x[p], x[q]], [y[p], y[q]], color='black')
            
            if cluster_labels:
                center_x = (x[p]+x[q])/2
                center_y = (y[p]+y[q])/2
                circle = plt.Circle((center_x, center_y), 0.1, color='white')
                ax.add_patch(circle)
                
                # Add text label
                ax.text(center_x, center_y, 'x_' + quiver.cluster_names[i], 
                        horizontalalignment='center',
                        verticalalignment='center', fontsize=fontsize)
        
        ############################ LAMINATIONS ############################
        if plot_laminations:
            for i, lamination in enumerate(quiver.laminations):
                for edge in lamination:
                    p, q = edge
                    
                    x1 = (x[p]+x[(p+1)%n])/2
                    y1 = (y[p]+y[(p+1)%n])/2
                    x2 = (x[q]+x[(q+1)%n])/2
                    y2 = (y[q]+y[(q+1)%n])/2
                    
                    ax.plot([x1, x2], [y1, y2], color=lamination_palette[i])
            
                if laminations_labels:
                    center_x = x1 + lamination_label_pos * (x2-x1)
                    center_y = y1 + lamination_label_pos * (y2-y1)
                    circle = plt.Circle((center_x, center_y), 0.1, color='white')
                    ax.add_patch(circle)
                    
                    # Add text label
                    ax.text(center_x, center_y, 'u_' + quiver.lamination_names[i], 
                            horizontalalignment='center',
                            verticalalignment='center', fontsize=fontsize)
        
        return ax


from setuptools import setup, find_packages

setup(
    name="cluster_algebra",
    version="0.1.1",
    packages=find_packages(),
    install_requires=[
        "numpy>=2.0.2",
        "pandas>=2.2.3",
        "matplotlib>=3.9.2",
        "networkx>=3.2.1",
        "scipy>=1.13.1",
        "sympy>=1.13.3",
        "shapely>=2.0.6",
        "pydantic>=2.9.2",
    ],
    extras_require={
        'dev': [
            'pytest>=8.3.3',
        ],
    },
    description="calculator for cluster algebras",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires='>=3.9',
)